@@ -6,6 +6,7 @@
 #include <graphics/matrix4.h>
 #include <obs-module.h>
 #include <QColorDialog>
+#include <QDesktopServices>
 #include <QFileDialog>
 #include <QGuiApplication>
 #include <QMainWindow>
@@ -35,18 +36,7 @@ bool obs_module_load()
 	const auto main_window = static_cast<QMainWindow *>(obs_frontend_get_main_window());
 	obs_frontend_push_ui_translation(obs_module_get_string);
 	draw_dock = new DrawDock(main_window);
-#if LIBOBS_API_VER >= MAKE_SEMANTIC_VERSION(30, 0, 0)
 	obs_frontend_add_dock_by_id("DrawDock", obs_module_text("DrawDock"), draw_dock);
-#else
-	const auto dock = new QDockWidget(main_window);
-	dock->setObjectName("DrawDock");
-	dock->setWindowTitle(QString::fromUtf8(obs_module_text("DrawDock")));
-	dock->setWidget(draw_dock);
-	dock->setFeatures(QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetFloatable);
-	dock->setFloating(true);
-	dock->hide();
-	obs_frontend_add_dock(dock);
-#endif
 	obs_frontend_pop_ui_translation();
 
 	return true;
@@ -164,6 +154,41 @@ static inline long long color_to_int(QColor color)
 	return shift(color.red(), 0) | shift(color.green(), 8) | shift(color.blue(), 16) | shift(color.alpha(), 24);
 }
 
+#ifdef _WIN32
+bool IsAlwaysOnTop(QWidget *window)
+{
+	DWORD exStyle = GetWindowLong((HWND)window->winId(), GWL_EXSTYLE);
+	return (exStyle & WS_EX_TOPMOST) != 0;
+}
+#else
+bool IsAlwaysOnTop(QWidget *window)
+{
+	return (window->windowFlags() & Qt::WindowStaysOnTopHint) != 0;
+}
+#endif
+
+#ifdef _WIN32
+void SetAlwaysOnTop(QWidget *window, bool enable)
+{
+	HWND hwnd = (HWND)window->winId();
+	SetWindowPos(hwnd, enable ? HWND_TOPMOST : HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
+}
+#else
+void SetAlwaysOnTop(QWidget *window, bool enable)
+{
+	Qt::WindowFlags flags = window->windowFlags();
+
+	if (enable) {
+		flags |= Qt::WindowStaysOnTopHint;
+	} else {
+		flags &= ~Qt::WindowStaysOnTopHint;
+	}
+
+	window->setWindowFlags(flags);
+	window->show();
+}
+#endif
+
 DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventFilter()), preview(new OBSQTDisplay(this))
 {
 	auto ml = new QVBoxLayout(this);
@@ -330,11 +355,23 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 			obs_source_update(draw_source, settings);
 			obs_data_release(settings);
 		});
+		a = cursorMenu->addAction(QString::fromUtf8(obs_module_text("CursorCustomSize")));
+		a->setCheckable(true);
+		a->setChecked(obs_data_get_bool(settings, "custom_cursor_size"));
+		connect(a, &QAction::triggered, [this, a] {
+			if (!draw_source)
+				return;
+			obs_data_t *settings = obs_data_create();
+			obs_data_set_bool(settings, "custom_cursor_size", a->isChecked());
+			obs_source_update(draw_source, settings);
+			obs_data_release(settings);
+		});
 		auto wa = new QWidgetAction(cursorMenu);
 		auto cursorSize = new QDoubleSpinBox();
 		cursorSize->setSuffix("px");
 		cursorSize->setValue(obs_data_get_double(settings, "cursor_size"));
 		cursorSize->setRange(0.0, 1000.0);
+		cursorSize->setEnabled(a->isChecked());
 		wa->setDefaultWidget(cursorSize);
 		cursorMenu->addAction(wa);
 
@@ -346,6 +383,19 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 			obs_source_update(draw_source, settings);
 			obs_data_release(settings);
 		});
+
+		a = menu.addAction(QString::fromUtf8(obs_module_text("ClearOnSceneTransition")));
+		a->setCheckable(true);
+		a->setChecked(obs_data_get_bool(settings, "clear_on_scene_transition"));
+		connect(a, &QAction::triggered, [this, a] {
+			if (!draw_source)
+				return;
+			obs_data_t *settings = obs_data_create();
+			obs_data_set_bool(settings, "clear_on_scene_transition", a->isChecked());
+			obs_source_update(draw_source, settings);
+			obs_data_release(settings);
+		});
+
 		menu.addSeparator();
 
 		menu.addAction(QString::fromUtf8(obs_module_text("Undo")), [this] {
@@ -360,24 +410,12 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 			if (!scene_source)
 				return;
 			obs_scene_t *scene = obs_scene_from_source(scene_source);
+			if (!scene)
+				scene = obs_group_from_source(scene_source);
 			obs_source_release(scene_source);
 			if (!scene)
 				return;
-
-			obs_scene_enum_items(
-				scene,
-				[](obs_scene_t *, obs_sceneitem_t *item, void *) {
-					auto source = obs_sceneitem_get_source(item);
-					if (!source || strcmp(obs_source_get_unversioned_id(source), "draw_source") != 0)
-						return true;
-					proc_handler_t *ph = obs_source_get_proc_handler(source);
-					if (!ph)
-						return true;
-					calldata_t cd = {};
-					proc_handler_call(ph, "undo", &cd);
-					return true;
-				},
-				nullptr);
+			obs_scene_enum_items(scene, scene_undo, nullptr);
 		});
 
 		menu.addAction(QString::fromUtf8(obs_module_text("Redo")), [this] {
@@ -392,24 +430,12 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 			if (!scene_source)
 				return;
 			obs_scene_t *scene = obs_scene_from_source(scene_source);
+			if (!scene)
+				scene = obs_group_from_source(scene_source);
 			obs_source_release(scene_source);
 			if (!scene)
 				return;
-
-			obs_scene_enum_items(
-				scene,
-				[](obs_scene_t *, obs_sceneitem_t *item, void *) {
-					auto source = obs_sceneitem_get_source(item);
-					if (!source || strcmp(obs_source_get_unversioned_id(source), "draw_source") != 0)
-						return true;
-					proc_handler_t *ph = obs_source_get_proc_handler(source);
-					if (!ph)
-						return true;
-					calldata_t cd = {};
-					proc_handler_call(ph, "redo", &cd);
-					return true;
-				},
-				nullptr);
+			obs_scene_enum_items(scene, scene_redo, nullptr);
 		});
 		auto undoMenu = menu.addMenu(QString::fromUtf8(obs_module_text("UndoMax")));
 		auto undowa = new QWidgetAction(undoMenu);
@@ -420,18 +446,21 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 		undoMenu->addAction(undowa);
 
 		connect(maxUndo, &QSpinBox::valueChanged, [this, maxUndo] {
-			if (!draw_source)
-				return;
-			obs_data_t *settings = obs_data_create();
-			obs_data_set_int(settings, "max_undo", maxUndo->value());
-			obs_source_update(draw_source, settings);
-			obs_data_release(settings);
+			if (draw_source) {
+				obs_data_t *settings = obs_data_create();
+				obs_data_set_int(settings, "max_undo", maxUndo->value());
+				obs_source_update(draw_source, settings);
+				obs_data_release(settings);
+			}
 		});
 
 		obs_data_release(settings);
 
 		menu.addSeparator();
-		auto fullMenu = menu.addMenu(QString::fromUtf8(obs_module_text("Fullscreen")));
+		auto d = (QDockWidget *)parent();
+		auto action = menu.addAction(QString::fromUtf8(obs_module_text("Fullscreen")));
+		auto fullMenu = new QMenu();
+		action->setMenu(fullMenu);
 		QList<QScreen *> screens = QGuiApplication::screens();
 		for (int i = 0; i < screens.size(); i++) {
 			QScreen *screen = screens[i];
@@ -467,7 +496,10 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 			QAction *a = fullMenu->addAction(str, this, SLOT(OpenFullScreenProjector()));
 			a->setProperty("monitor", i);
 		}
-		menu.addAction(QString::fromUtf8(obs_module_text("Dock")), [this] {
+		action->setCheckable(true);
+		action->setChecked(d->parent() == nullptr && config && obs_data_get_bool(config, "fullscreen"));
+
+		action = menu.addAction(QString::fromUtf8(obs_module_text("Dock")), [this] {
 			auto dock = (QDockWidget *)parent();
 			auto main = static_cast<QMainWindow *>(obs_frontend_get_main_window());
 			if (!dock->parent()) {
@@ -499,7 +531,9 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 				obs_data_set_bool(config, "windowed", false);
 			}
 		});
-		menu.addAction(QString::fromUtf8(obs_module_text("Windowed")), [this] {
+		action->setCheckable(true);
+		action->setChecked(d->parent() != nullptr);
+		action = menu.addAction(QString::fromUtf8(obs_module_text("Windowed")), [this] {
 			auto dock = (QDockWidget *)parent();
 			if (dock->parent()) {
 				prevGeometry = dock->geometry();
@@ -518,18 +552,42 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 				obs_data_set_bool(config, "windowed", true);
 			}
 		});
+		action->setCheckable(true);
+		action->setChecked(d->parent() == nullptr && (!config || obs_data_get_bool(config, "windowed")));
+		action = menu.addAction(QString::fromUtf8(obs_module_text("AlwaysOnTop")), [&] {
+			auto dock = (QDockWidget *)parent();
+			bool aot = !IsAlwaysOnTop(dock);
+			SetAlwaysOnTop(dock, aot);
+			if (config)
+				obs_data_set_bool(config, "always_on_top", aot);
+		});
+		action->setCheckable(true);
+		action->setChecked(IsAlwaysOnTop((QDockWidget *)parent()));
+
+		menu.addSeparator();
+		menu.addAction(QString::fromUtf8("Draw (" PROJECT_VERSION ")"),
+			       [] { QDesktopServices::openUrl(QUrl("https://obsproject.com/forum/resources/draw.2081/")); });
+		menu.addAction(QString::fromUtf8("By Exeldro"), [] { QDesktopServices::openUrl(QUrl("https://www.exeldro.com")); });
 
 		menu.exec(QCursor::pos());
 	});
 	toolbar->widgetForAction(a)->setProperty("themeID", "propertiesIconSmall");
 	toolbar->widgetForAction(a)->setProperty("class", "icon-gear");
 
-	auto hotkeyId = obs_hotkey_register_frontend("draw_clear", obs_module_text("DrawClear"), clear_hotkey, this);
+	clearHotkey = obs_hotkey_register_frontend("draw_clear", obs_module_text("DrawClear"), clear_hotkey, this);
 	auto hotkeys = obs_data_get_array(config, "clear_hotkey");
 	if (hotkeys) {
-		obs_hotkey_load(hotkeyId, hotkeys);
+		obs_hotkey_load(clearHotkey, hotkeys);
 		obs_data_array_release(hotkeys);
 	}
+	showHideHotkey = obs_hotkey_pair_register_frontend("draw_show", obs_module_text("DrawShow"), "draw_hide",
+							   obs_module_text("DrawHide"), show_hotkey, hide_hotkey, this, this);
+
+	obs_data_array_t *show_hotkeys = obs_data_get_array(config, "show_hotkey");
+	obs_data_array_t *hide_hotkeys = obs_data_get_array(config, "hide_hotkey");
+	obs_hotkey_pair_load(showHideHotkey, show_hotkeys, hide_hotkeys);
+	obs_data_array_release(show_hotkeys);
+	obs_data_array_release(hide_hotkeys);
 
 	obs_data_array_t *tools = obs_data_get_array(config, "tools");
 	auto count = obs_data_array_count(tools);
@@ -589,22 +647,7 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 		if (!scene)
 			return;
 
-		obs_scene_enum_items(
-			scene,
-			[](obs_scene_t *, obs_sceneitem_t *item, void *data) {
-				auto source = obs_sceneitem_get_source(item);
-				if (!source || strcmp(obs_source_get_unversioned_id(source), "draw_source") != 0)
-					return true;
-				int tool = *((int *)data);
-				obs_data_t *ss = obs_source_get_settings(source);
-				if (obs_data_get_int(ss, "tool") != tool) {
-					obs_data_set_int(ss, "tool", tool);
-					obs_source_update(source, ss);
-				}
-				obs_data_release(ss);
-				return true;
-			},
-			&tool);
+		obs_scene_enum_items(scene, scene_tool, &tool);
 	});
 	toolbar->addWidget(toolCombo);
 	colorAction = toolbar->addAction(QString::fromUtf8(obs_module_text("ToolColor")), [this] {
@@ -633,22 +676,7 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 		if (!scene)
 			return;
 
-		obs_scene_enum_items(
-			scene,
-			[](obs_scene_t *, obs_sceneitem_t *item, void *data) {
-				auto source = obs_sceneitem_get_source(item);
-				if (!source || strcmp(obs_source_get_unversioned_id(source), "draw_source") != 0)
-					return true;
-				long long longColor = *((long long *)data);
-				obs_data_t *ss = obs_source_get_settings(source);
-				if (obs_data_get_int(ss, "tool_color") != longColor) {
-					obs_data_set_int(ss, "tool_color", longColor);
-					obs_source_update(source, ss);
-				}
-				obs_data_release(ss);
-				return true;
-			},
-			&longColor);
+		obs_scene_enum_items(scene, scene_tool_color, &longColor);
 	});
 	imageAction = toolbar->addAction(QString::fromUtf8(obs_module_text("ToolImage")), [this] {
 		if (!draw_source)
@@ -673,22 +701,7 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 		obs_source_release(scene_source);
 		if (!scene)
 			return;
-		obs_scene_enum_items(
-			scene,
-			[](obs_scene_t *, obs_sceneitem_t *item, void *data) {
-				auto source = obs_sceneitem_get_source(item);
-				if (!source || strcmp(obs_source_get_unversioned_id(source), "draw_source") != 0)
-					return true;
-				const char *path = (const char *)data;
-				obs_data_t *ss = obs_source_get_settings(source);
-				if (strcmp(obs_data_get_string(ss, "tool_image_file"), path) != 0) {
-					obs_data_set_string(ss, "tool_image_file", path);
-					obs_source_update(source, ss);
-				}
-				obs_data_release(ss);
-				return true;
-			},
-			(void *)fileName.toUtf8().constData());
+		obs_scene_enum_items(scene, scene_tool_image, (void *)fileName.toUtf8().constData());
 	});
 	imageAction->setVisible(false);
 	toolSizeSpin = new QDoubleSpinBox;
@@ -714,22 +727,7 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 		if (!scene)
 			return;
 
-		obs_scene_enum_items(
-			scene,
-			[](obs_scene_t *, obs_sceneitem_t *item, void *data) {
-				auto source = obs_sceneitem_get_source(item);
-				if (!source || strcmp(obs_source_get_unversioned_id(source), "draw_source") != 0)
-					return true;
-				double size = *((double *)data);
-				obs_data_t *ss = obs_source_get_settings(source);
-				if (abs(obs_data_get_double(ss, "tool_size") - size) > 0.1) {
-					obs_data_set_double(ss, "tool_size", size);
-					obs_source_update(source, ss);
-				}
-				obs_data_release(ss);
-				return true;
-			},
-			&size);
+		obs_scene_enum_items(scene, scene_tool_size, &size);
 	});
 
 	toolbar->addWidget(toolSizeSpin);
@@ -763,22 +761,7 @@ DrawDock::DrawDock(QWidget *_parent) : QWidget(_parent), eventFilter(BuildEventF
 		if (!scene)
 			return;
 
-		obs_scene_enum_items(
-			scene,
-			[](obs_scene_t *, obs_sceneitem_t *item, void *data) {
-				auto source = obs_sceneitem_get_source(item);
-				if (!source || strcmp(obs_source_get_unversioned_id(source), "draw_source") != 0)
-					return true;
-				double alpha = *((double *)data);
-				obs_data_t *ss = obs_source_get_settings(source);
-				if (abs(obs_data_get_double(ss, "tool_alpha") - alpha) > 0.1) {
-					obs_data_set_double(ss, "tool_alpha", alpha);
-					obs_source_update(source, ss);
-				}
-				obs_data_release(ss);
-				return true;
-			},
-			&alpha);
+		obs_scene_enum_items(scene, scene_tool_alpha, &alpha);
 	};
 
 	connect(alphaSpin, &QDoubleSpinBox::valueChanged, alphaChange);
@@ -821,6 +804,8 @@ DrawDock::~DrawDock()
 {
 	if (clearHotkey != OBS_INVALID_HOTKEY_ID)
 		obs_hotkey_unregister(clearHotkey);
+	if (showHideHotkey != OBS_INVALID_HOTKEY_PAIR_ID)
+		obs_hotkey_pair_unregister(showHideHotkey);
 	for (auto i = favoriteToolHotkeys.begin(); i != favoriteToolHotkeys.end(); i++) {
 		obs_hotkey_unregister(i->first);
 	}
@@ -950,6 +935,7 @@ bool DrawDock::GetSourceRelativeXY(int mouseX, int mouseY, int &relX, int &relY)
 
 	return true;
 }
+
 static int TranslateQtKeyboardEventModifiers(QInputEvent *event, bool mouseEvent)
 {
 	int obsModifiers = INTERACT_NONE;
@@ -1015,7 +1001,8 @@ static bool HandleSceneMouseClickEvent(obs_scene_t *scene, obs_sceneitem_t *item
 	if (!obs_sceneitem_visible(item))
 		return true;
 	auto source = obs_sceneitem_get_source(item);
-	if (!source || strcmp(obs_source_get_unversioned_id(source), "draw_source") != 0)
+	if (!source || (strcmp(obs_source_get_unversioned_id(source), "draw_source") != 0 && !obs_source_is_group(source) &&
+			!obs_source_is_scene(source)))
 		return true;
 
 	auto click_event = static_cast<struct click_event *>(data);
@@ -1036,11 +1023,33 @@ static bool HandleSceneMouseClickEvent(obs_scene_t *scene, obs_sceneitem_t *item
 
 	if (CloseFloat(pos3.x, pos3_.x) && CloseFloat(pos3.y, pos3_.y) && transformedPos.x >= 0.0f && transformedPos.x <= 1.0f &&
 	    transformedPos.y >= 0.0f && transformedPos.y <= 1.0f) {
-		click_event->mouseEvent.x = transformedPos.x * obs_source_get_base_width(source);
-		click_event->mouseEvent.y = transformedPos.y * obs_source_get_base_height(source);
-		click_event->mouseEvent.modifiers = click_event->modifiers;
-		click_event->mouseTarget = source;
-		return false;
+		if (obs_source_is_group(source) || obs_source_is_scene(source)) {
+			obs_scene_t *ss = obs_scene_from_source(source);
+			if (!ss)
+				ss = obs_group_from_source(source);
+
+			struct click_event ce{(int32_t)(transformedPos.x * obs_source_get_base_width(source)),
+					      (int32_t)(transformedPos.y * obs_source_get_base_height(source)),
+					      click_event->modifiers,
+					      click_event->button,
+					      click_event->mouseUp,
+					      click_event->clickCount,
+					      nullptr};
+			obs_scene_enum_items(ss, HandleSceneMouseClickEvent, &ce);
+			if (ce.mouseTarget) {
+				click_event->mouseEvent.x = ce.mouseEvent.x;
+				click_event->mouseEvent.y = ce.mouseEvent.y;
+				click_event->mouseEvent.modifiers = ce.modifiers;
+				click_event->mouseTarget = ce.mouseTarget;
+				return false;
+			}
+		} else {
+			click_event->mouseEvent.x = transformedPos.x * obs_source_get_base_width(source);
+			click_event->mouseEvent.y = transformedPos.y * obs_source_get_base_height(source);
+			click_event->mouseEvent.modifiers = click_event->modifiers;
+			click_event->mouseTarget = source;
+			return false;
+		}
 	}
 
 	return true;
@@ -1150,7 +1159,8 @@ static bool HandleSceneMouseMoveEvent(obs_scene_t *scene, obs_sceneitem_t *item,
 	if (!obs_sceneitem_visible(item))
 		return true;
 	obs_source_t *source = obs_sceneitem_get_source(item);
-	if (!source || strcmp(obs_source_get_unversioned_id(source), "draw_source") != 0)
+	if (!source || (strcmp(obs_source_get_unversioned_id(source), "draw_source") != 0 && !obs_source_is_group(source) &&
+			!obs_source_is_scene(source)))
 		return true;
 
 	auto move_event = static_cast<struct move_event *>(data);
@@ -1171,11 +1181,30 @@ static bool HandleSceneMouseMoveEvent(obs_scene_t *scene, obs_sceneitem_t *item,
 
 	if (CloseFloat(pos3.x, pos3_.x) && CloseFloat(pos3.y, pos3_.y) && transformedPos.x >= 0.0f && transformedPos.x <= 1.0f &&
 	    transformedPos.y >= 0.0f && transformedPos.y <= 1.0f) {
-		move_event->mouseEvent.x = transformedPos.x * obs_source_get_base_width(source);
-		move_event->mouseEvent.y = transformedPos.y * obs_source_get_base_height(source);
-		move_event->mouseEvent.modifiers = move_event->modifiers;
-		move_event->mouseTarget = source;
-		return false;
+
+		if (obs_source_is_group(source) || obs_source_is_scene(source)) {
+			obs_scene_t *ss = obs_scene_from_source(source);
+			if (!ss)
+				ss = obs_group_from_source(source);
+
+			struct move_event ce{(int32_t)(transformedPos.x * obs_source_get_base_width(source)),
+					     (int32_t)(transformedPos.y * obs_source_get_base_height(source)),
+					     move_event->modifiers, move_event->mouseLeave, nullptr};
+			obs_scene_enum_items(ss, HandleSceneMouseMoveEvent, &ce);
+			if (ce.mouseTarget) {
+				move_event->mouseEvent.x = ce.mouseEvent.x;
+				move_event->mouseEvent.y = ce.mouseEvent.y;
+				move_event->mouseEvent.modifiers = ce.modifiers;
+				move_event->mouseTarget = ce.mouseTarget;
+				return false;
+			}
+		} else {
+			move_event->mouseEvent.x = transformedPos.x * obs_source_get_base_width(source);
+			move_event->mouseEvent.y = transformedPos.y * obs_source_get_base_height(source);
+			move_event->mouseEvent.modifiers = move_event->modifiers;
+			move_event->mouseTarget = source;
+			return false;
+		}
 	}
 
 	obs_mouse_event mouseEvent;
@@ -1489,14 +1518,15 @@ void DrawDock::frontend_event(enum obs_frontend_event event, void *data)
 	} else if (event == OBS_FRONTEND_EVENT_SCENE_COLLECTION_CLEANUP || event == OBS_FRONTEND_EVENT_EXIT ||
 		   event == OBS_FRONTEND_EVENT_SCRIPTING_SHUTDOWN || event == OBS_FRONTEND_EVENT_SCENE_COLLECTION_CHANGING) {
 		window->DestroyDrawSource();
-	} else if (event == OBS_FRONTEND_EVENT_SCENE_CHANGED || event == OBS_FRONTEND_EVENT_STUDIO_MODE_ENABLED ||
-		   event == OBS_FRONTEND_EVENT_STUDIO_MODE_DISABLED || event == OBS_FRONTEND_EVENT_PREVIEW_SCENE_CHANGED) {
-		QMetaObject::invokeMethod(window, "SceneChanged", Qt::QueuedConnection);
 	}
 }
 
 void DrawDock::CreateDrawSource(obs_source_t *new_source)
 {
+	static bool loading = false;
+	if (loading)
+		return;
+	loading = true;
 	bool set_output = true;
 	for (uint32_t i = MAX_CHANNELS - 1; i > 0; i--) {
 		obs_source_t *source = obs_get_output_source(i);
@@ -1525,6 +1555,7 @@ void DrawDock::CreateDrawSource(obs_source_t *new_source)
 	if (draw_source && strcmp(obs_source_get_unversioned_id(draw_source), "draw_source") != 0) {
 		obs_source_release(draw_source);
 		draw_source = nullptr;
+		loading = false;
 		return;
 	}
 
@@ -1559,6 +1590,8 @@ void DrawDock::CreateDrawSource(obs_source_t *new_source)
 	}
 	obs_data_release(settings);
 
+	loading = false;
+
 	signal_handler_t *sh = obs_source_get_signal_handler(draw_source);
 	signal_handler_connect(sh, "update", draw_source_update, this);
 	signal_handler_connect(sh, "destroy", draw_source_destroy, this);
@@ -1566,6 +1599,10 @@ void DrawDock::CreateDrawSource(obs_source_t *new_source)
 		for (uint32_t i = MAX_CHANNELS - 1; i > 0; i--) {
 			obs_source_t *source = obs_get_output_source(i);
 			if (source) {
+				if (strcmp(obs_source_get_unversioned_id(source), "draw_source") == 0) {
+					obs_source_release(source);
+					return;
+				}
 				obs_source_release(source);
 				continue;
 			}
@@ -1622,7 +1659,9 @@ void DrawDock::DestroyDrawSource()
 
 	for (uint32_t i = 0; i < MAX_CHANNELS; i++) {
 		obs_source_t *s = obs_get_output_source(i);
-		if (s == source) {
+		if (!s)
+			continue;
+		if (s == source || strcmp(obs_source_get_name(s), "Global Draw Source") == 0) {
 			obs_set_output_source(i, nullptr);
 		}
 		obs_source_release(s);
@@ -1639,8 +1678,22 @@ void DrawDock::SaveConfig()
 	ensure_directory(path);
 
 	obs_data_array_t *clearHotkeyData = obs_hotkey_save(clearHotkey);
-	obs_data_set_array(config, "clear_hotkey", clearHotkeyData);
-	obs_data_array_release(clearHotkeyData);
+	if (clearHotkeyData) {
+		obs_data_set_array(config, "clear_hotkey", clearHotkeyData);
+		obs_data_array_release(clearHotkeyData);
+	}
+
+	obs_data_array_t *showHotkeyData = nullptr;
+	obs_data_array_t *hideHotkeyData = nullptr;
+	obs_hotkey_pair_save(showHideHotkey, &showHotkeyData, &hideHotkeyData);
+	if (showHotkeyData) {
+		obs_data_set_array(config, "show_hotkey", showHotkeyData);
+		obs_data_array_release(showHotkeyData);
+	}
+	if (hideHotkeyData) {
+		obs_data_set_array(config, "hide_hotkey", hideHotkeyData);
+		obs_data_array_release(hideHotkeyData);
+	}
 
 	obs_data_array_t *tools = obs_data_get_array(config, "tools");
 	size_t count = obs_data_array_count(tools);
@@ -1660,6 +1713,11 @@ void DrawDock::SaveConfig()
 	}
 	obs_data_array_release(tools);
 
+	if (obs_data_get_bool(config, "windowed")) {
+		auto dock = (QDockWidget *)parent();
+		obs_data_set_string(config, "window_geometry", dock->saveGeometry().toBase64().constData());
+	}
+
 	if (obs_data_save_json_safe(config, path, "tmp", "bak")) {
 		blog(LOG_INFO, "[Draw Dock] Saved settings");
 	} else {
@@ -1838,6 +1896,46 @@ void DrawDock::clear_hotkey(void *data, obs_hotkey_id id, obs_hotkey_t *hotkey,
 	window->ClearDraw();
 }
 
+bool DrawDock::show_hotkey(void *data, obs_hotkey_pair_id id, obs_hotkey_t *hotkey, bool pressed)
+{
+	UNUSED_PARAMETER(hotkey);
+	UNUSED_PARAMETER(id);
+	if (!pressed)
+		return false;
+	DrawDock *window = static_cast<DrawDock *>(data);
+	if (window->parentWidget()->isHidden()) {
+		obs_queue_task(
+			OBS_TASK_UI,
+			[](void *param) {
+				DrawDock *window = static_cast<DrawDock *>(param);
+				window->parentWidget()->show();
+			},
+			data, false);
+		return true;
+	}
+	return false;
+}
+
+bool DrawDock::hide_hotkey(void *data, obs_hotkey_pair_id id, obs_hotkey_t *hotkey, bool pressed)
+{
+	UNUSED_PARAMETER(hotkey);
+	UNUSED_PARAMETER(id);
+	if (!pressed)
+		return false;
+	DrawDock *window = static_cast<DrawDock *>(data);
+	if (!window->parentWidget()->isHidden()) {
+		obs_queue_task(
+			OBS_TASK_UI,
+			[](void *param) {
+				DrawDock *window = static_cast<DrawDock *>(param);
+				window->parentWidget()->hide();
+			},
+			data, false);
+		return true;
+	}
+	return false;
+}
+
 void DrawDock::favorite_tool_hotkey(void *data, obs_hotkey_id id, obs_hotkey_t *hotkey, bool pressed)
 {
 	UNUSED_PARAMETER(hotkey);
@@ -1867,17 +1965,7 @@ void DrawDock::ApplyFavoriteTool(obs_data_t *settings)
 	if (!scene)
 		return;
 
-	obs_scene_enum_items(
-		scene,
-		[](obs_scene_t *, obs_sceneitem_t *item, void *data) {
-			auto source = obs_sceneitem_get_source(item);
-			if (!source || strcmp(obs_source_get_unversioned_id(source), "draw_source") != 0)
-				return true;
-			obs_data_t *settings = (obs_data_t *)data;
-			obs_source_update(source, settings);
-			return true;
-		},
-		settings);
+	obs_scene_enum_items(scene, scene_apply_tool, settings);
 }
 
 QIcon DrawDock::CreateToolIcon(QColor toolColor, uint32_t tool, double alpha, double toolSize, const char *image)
@@ -1975,9 +2063,8 @@ void DrawDock::FinishedLoad()
 	imageAction->setIcon(imageIcon);
 	toolCombo->setItemIcon(TOOL_STAMP, imageIcon);
 	toolCombo->setItemIcon(TOOL_IMAGE, imageIcon);
+	auto dock = (QDockWidget *)parent();
 	if (obs_data_get_bool(config, "fullscreen")) {
-
-		auto dock = (QDockWidget *)parent();
 		dock->setFloating(true);
 		dock->setParent(nullptr);
 		dock->setGeometry(QRect(obs_data_get_int(config, "fullscreen_left"), obs_data_get_int(config, "fullscreen_top"),
@@ -1985,11 +2072,18 @@ void DrawDock::FinishedLoad()
 					obs_data_get_int(config, "fullscreen_height")));
 		dock->showFullScreen();
 	} else if (obs_data_get_bool(config, "windowed")) {
-		auto dock = (QDockWidget *)parent();
 		dock->setFloating(true);
 		dock->setParent(nullptr);
 		dock->showNormal();
+
+		const char *geom = obs_data_get_string(config, "window_geometry");
+		if (geom && strlen(geom)) {
+			QByteArray ba = QByteArray::fromBase64(QByteArray(geom));
+			dock->restoreGeometry(ba);
+		}
 	}
+	if (obs_data_get_bool(config, "always_on_top"))
+		SetAlwaysOnTop(dock, true);
 }
 
 void DrawDock::vendor_request_version(obs_data_t *request_data, obs_data_t *response_data, void *)
@@ -2150,3 +2244,198 @@ void DrawDock::EscapeTriggered()
 		dock->resize(860, 530);
 	}
 }
+
+void DrawDock::showEvent(QShowEvent *)
+{
+	if (vendor)
+		obs_websocket_vendor_emit_event(vendor, "show", nullptr);
+}
+
+void DrawDock::hideEvent(QHideEvent *)
+{
+	if (vendor)
+		obs_websocket_vendor_emit_event(vendor, "hide", nullptr);
+}
+
+bool DrawDock::scene_undo(obs_scene_t *, obs_sceneitem_t *item, void *data)
+{
+	if (!obs_sceneitem_visible(item))
+		return true;
+	auto source = obs_sceneitem_get_source(item);
+	if (!source)
+		return true;
+	obs_scene_t *scene = obs_scene_from_source(source);
+	if (!scene)
+		scene = obs_group_from_source(source);
+	if (scene) {
+		obs_scene_enum_items(scene, scene_undo, data);
+	} else if (strcmp(obs_source_get_unversioned_id(source), "draw_source") == 0) {
+		proc_handler_t *ph = obs_source_get_proc_handler(source);
+		if (!ph)
+			return true;
+		calldata_t cd = {};
+		proc_handler_call(ph, "undo", &cd);
+	}
+	return true;
+}
+
+bool DrawDock::scene_redo(obs_scene_t *, obs_sceneitem_t *item, void *data)
+{
+	if (!obs_sceneitem_visible(item))
+		return true;
+	auto source = obs_sceneitem_get_source(item);
+	if (!source)
+		return true;
+	obs_scene_t *scene = obs_scene_from_source(source);
+	if (!scene)
+		scene = obs_group_from_source(source);
+	if (scene) {
+		obs_scene_enum_items(scene, scene_redo, data);
+	} else if (strcmp(obs_source_get_unversioned_id(source), "draw_source") == 0) {
+		proc_handler_t *ph = obs_source_get_proc_handler(source);
+		if (!ph)
+			return true;
+		calldata_t cd = {};
+		proc_handler_call(ph, "redo", &cd);
+	}
+	return true;
+}
+
+bool DrawDock::scene_tool(obs_scene_t *, obs_sceneitem_t *item, void *data)
+{
+	if (!obs_sceneitem_visible(item))
+		return true;
+	auto source = obs_sceneitem_get_source(item);
+	if (!source)
+		return true;
+	obs_scene_t *scene = obs_scene_from_source(source);
+	if (!scene)
+		scene = obs_group_from_source(source);
+	if (scene) {
+		obs_scene_enum_items(scene, scene_tool, data);
+	} else if (strcmp(obs_source_get_unversioned_id(source), "draw_source") == 0) {
+		int tool = *((int *)data);
+		obs_data_t *ss = obs_source_get_settings(source);
+		if (obs_data_get_int(ss, "tool") != tool) {
+			obs_data_set_int(ss, "tool", tool);
+			obs_source_update(source, ss);
+		}
+		obs_data_release(ss);
+	}
+	return true;
+}
+
+bool DrawDock::scene_tool_color(obs_scene_t *, obs_sceneitem_t *item, void *data)
+{
+	if (!obs_sceneitem_visible(item))
+		return true;
+	auto source = obs_sceneitem_get_source(item);
+	if (!source)
+		return true;
+	obs_scene_t *scene = obs_scene_from_source(source);
+	if (!scene)
+		scene = obs_group_from_source(source);
+	if (scene) {
+		obs_scene_enum_items(scene, scene_tool_color, data);
+	} else if (strcmp(obs_source_get_unversioned_id(source), "draw_source") == 0) {
+		long long longColor = *((long long *)data);
+		obs_data_t *ss = obs_source_get_settings(source);
+		if (obs_data_get_int(ss, "tool_color") != longColor) {
+			obs_data_set_int(ss, "tool_color", longColor);
+			obs_source_update(source, ss);
+		}
+		obs_data_release(ss);
+	}
+	return true;
+}
+
+bool DrawDock::scene_tool_image(obs_scene_t *, obs_sceneitem_t *item, void *data)
+{
+	if (!obs_sceneitem_visible(item))
+		return true;
+	auto source = obs_sceneitem_get_source(item);
+	if (!source)
+		return true;
+	obs_scene_t *scene = obs_scene_from_source(source);
+	if (!scene)
+		scene = obs_group_from_source(source);
+	if (scene) {
+		obs_scene_enum_items(scene, scene_tool_image, data);
+	} else if (strcmp(obs_source_get_unversioned_id(source), "draw_source") == 0) {
+		const char *path = (const char *)data;
+		obs_data_t *ss = obs_source_get_settings(source);
+		if (strcmp(obs_data_get_string(ss, "tool_image_file"), path) != 0) {
+			obs_data_set_string(ss, "tool_image_file", path);
+			obs_source_update(source, ss);
+		}
+		obs_data_release(ss);
+	}
+	return true;
+}
+
+bool DrawDock::scene_tool_size(obs_scene_t *, obs_sceneitem_t *item, void *data)
+{
+	if (!obs_sceneitem_visible(item))
+		return true;
+	auto source = obs_sceneitem_get_source(item);
+	if (!source)
+		return true;
+	obs_scene_t *scene = obs_scene_from_source(source);
+	if (!scene)
+		scene = obs_group_from_source(source);
+	if (scene) {
+		obs_scene_enum_items(scene, scene_tool_size, data);
+	} else if (strcmp(obs_source_get_unversioned_id(source), "draw_source") == 0) {
+		double size = *((double *)data);
+		obs_data_t *ss = obs_source_get_settings(source);
+		if (abs(obs_data_get_double(ss, "tool_size") - size) > 0.1) {
+			obs_data_set_double(ss, "tool_size", size);
+			obs_source_update(source, ss);
+		}
+		obs_data_release(ss);
+	}
+	return true;
+}
+
+bool DrawDock::scene_tool_alpha(obs_scene_t *, obs_sceneitem_t *item, void *data)
+{
+	if (!obs_sceneitem_visible(item))
+		return true;
+	auto source = obs_sceneitem_get_source(item);
+	if (!source)
+		return true;
+	obs_scene_t *scene = obs_scene_from_source(source);
+	if (!scene)
+		scene = obs_group_from_source(source);
+	if (scene) {
+		obs_scene_enum_items(scene, scene_tool_alpha, data);
+	} else if (strcmp(obs_source_get_unversioned_id(source), "draw_source") == 0) {
+		double alpha = *((double *)data);
+		obs_data_t *ss = obs_source_get_settings(source);
+		if (abs(obs_data_get_double(ss, "tool_alpha") - alpha) > 0.1) {
+			obs_data_set_double(ss, "tool_alpha", alpha);
+			obs_source_update(source, ss);
+		}
+		obs_data_release(ss);
+	}
+	return true;
+}
+
+bool DrawDock::scene_apply_tool(obs_scene_t *, obs_sceneitem_t *item, void *data)
+{
+	if (!obs_sceneitem_visible(item))
+		return true;
+	auto source = obs_sceneitem_get_source(item);
+	if (!source)
+		return true;
+	obs_scene_t *scene = obs_scene_from_source(source);
+	if (!scene)
+		scene = obs_group_from_source(source);
+	if (scene) {
+		obs_scene_enum_items(scene, scene_apply_tool, data);
+	} else if (strcmp(obs_source_get_unversioned_id(source), "draw_source") == 0) {
+		obs_data_t *settings = (obs_data_t *)data;
+		obs_source_update(source, settings);
+	}
+	return true;
+}
