@@ -1,35 +1,18 @@
 #include "draw-source.h"
 #include "version.h"
 #include <graphics/image-file.h>
+#include <obs-frontend-api.h>
 #include <obs-module.h>
-#if LIBOBS_API_VER >= MAKE_SEMANTIC_VERSION(30, 1, 0)
 #include <util/deque.h>
-#define circlebuf_peek_front deque_peek_front
-#define circlebuf_peek_back deque_peek_back
-#define circlebuf_push_front deque_push_front
-#define circlebuf_push_back deque_push_back
-#define circlebuf_pop_front deque_pop_front
-#define circlebuf_pop_back deque_pop_back
-#define circlebuf_init deque_init
-#define circlebuf_free deque_free
-#else
-#include <util/circlebuf.h>
-#endif
+
 
 struct draw_source {
 	obs_source_t *source;
 	struct vec2 size;
 
-#if LIBOBS_API_VER >= MAKE_SEMANTIC_VERSION(30, 1, 0)
 	struct deque undo;
-#else
-	struct circlebuf undo;
-#endif
-#if LIBOBS_API_VER >= MAKE_SEMANTIC_VERSION(30, 1, 0)
 	struct deque redo;
-#else
-	struct circlebuf redo;
-#endif
+
 	uint32_t max_undo;
 	gs_texrender_t *render_a;
 	gs_texrender_t *render_b;
@@ -76,6 +59,7 @@ struct draw_source {
 	char *cursor_image_path;
 	gs_image_file4_t *cursor_image;
 	uint64_t last_tick;
+	bool clear_on_transition;
 };
 
 const char *ds_get_name(void *data)
@@ -111,7 +95,7 @@ static void copy_to_undo(struct draw_source *ds)
 	obs_enter_graphics();
 	while (ds->redo.size) {
 		gs_texrender_t *old;
-		circlebuf_pop_front(&ds->redo, &old, sizeof(old));
+		deque_pop_front(&ds->redo, &old, sizeof(old));
 		gs_texrender_destroy(old);
 	}
 	gs_texrender_t *texrender = gs_texrender_create(GS_RGBA, GS_ZS_NONE);
@@ -126,16 +110,16 @@ static void copy_to_undo(struct draw_source *ds)
 			draw_effect(ds, tex, false);
 		gs_blend_state_pop();
 		gs_texrender_end(texrender);
-		circlebuf_push_back(&ds->undo, &texrender, sizeof(texrender));
+		deque_push_back(&ds->undo, &texrender, sizeof(texrender));
 		if (ds->undo.size > sizeof(texrender) * ds->max_undo) {
-			circlebuf_pop_front(&ds->undo, &texrender, sizeof(texrender));
+			deque_pop_front(&ds->undo, &texrender, sizeof(texrender));
 			gs_texrender_destroy(texrender);
 		}
 	}
 	obs_leave_graphics();
 }
 
-void clear(struct draw_source *ds)
+void draw_clear(struct draw_source *ds)
 {
 	copy_to_undo(ds);
 	obs_enter_graphics();
@@ -154,7 +138,7 @@ void clear_proc_handler(void *data, calldata_t *cd)
 {
 	UNUSED_PARAMETER(cd);
 	struct draw_source *context = data;
-	clear(context);
+	draw_clear(context);
 }
 
 static void apply_tool(struct draw_source *ds);
@@ -195,16 +179,16 @@ void undo(struct draw_source *ds)
 		return;
 
 	gs_texrender_t *texrender;
-	circlebuf_pop_back(&ds->undo, &texrender, sizeof(texrender));
+	deque_pop_back(&ds->undo, &texrender, sizeof(texrender));
 
 	if (ds->render_a_active) {
 		gs_texrender_t *old = ds->render_a;
 		ds->render_a = texrender;
-		circlebuf_push_back(&ds->redo, &old, sizeof(old));
+		deque_push_back(&ds->redo, &old, sizeof(old));
 	} else {
 		gs_texrender_t *old = ds->render_b;
 		ds->render_b = texrender;
-		circlebuf_push_back(&ds->redo, &old, sizeof(old));
+		deque_push_back(&ds->redo, &old, sizeof(old));
 	}
 }
 
@@ -221,16 +205,16 @@ void redo(struct draw_source *ds)
 		return;
 
 	gs_texrender_t *texrender = NULL;
-	circlebuf_pop_back(&ds->redo, &texrender, sizeof(texrender));
+	deque_pop_back(&ds->redo, &texrender, sizeof(texrender));
 
 	if (ds->render_a_active) {
 		gs_texrender_t *old = ds->render_a;
 		ds->render_a = texrender;
-		circlebuf_push_back(&ds->undo, &old, sizeof(old));
+		deque_push_back(&ds->undo, &old, sizeof(old));
 	} else {
 		gs_texrender_t *old = ds->render_b;
 		ds->render_b = texrender;
-		circlebuf_push_back(&ds->undo, &old, sizeof(old));
+		deque_push_back(&ds->undo, &old, sizeof(old));
 	}
 }
 
@@ -358,9 +342,20 @@ static void *ds_create(obs_data_t *settings, obs_source_t *source)
 	return context;
 }
 
+static void ds_frontend_event(enum obs_frontend_event event, void *data)
+{
+	struct draw_source *context = data;
+	if (event == OBS_FRONTEND_EVENT_SCENE_CHANGED) {
+		if (context->clear_on_transition) {
+			draw_clear(context);
+		}
+	}
+}
+
 static void ds_destroy(void *data)
 {
 	struct draw_source *context = data;
+	obs_frontend_remove_event_callback(ds_frontend_event, data);
 	bool graphics = false;
 	if (context->undo.size) {
 		graphics = true;
@@ -368,16 +363,16 @@ static void ds_destroy(void *data)
 	}
 	while (context->undo.size) {
 		gs_texrender_t *texrender;
-		circlebuf_pop_front(&context->undo, &texrender, sizeof(texrender));
+		deque_pop_front(&context->undo, &texrender, sizeof(texrender));
 		gs_texrender_destroy(texrender);
 	}
-	circlebuf_free(&context->undo);
+	deque_free(&context->undo);
 	while (context->redo.size) {
 		gs_texrender_t *texrender;
-		circlebuf_pop_front(&context->redo, &texrender, sizeof(texrender));
+		deque_pop_front(&context->redo, &texrender, sizeof(texrender));
 		gs_texrender_destroy(texrender);
 	}
-	circlebuf_free(&context->redo);
+	deque_free(&context->redo);
 	if (context->render_a) {
 		if (!graphics) {
 			graphics = true;
@@ -559,12 +554,21 @@ void ds_key_click(void *data, const struct obs_key_event *event, bool key_up)
 static void ds_update(void *data, obs_data_t *settings)
 {
 	struct draw_source *context = data;
+
+	bool clear_on_transition = obs_data_get_bool(settings, "clear_on_scene_transition");
+	if (clear_on_transition && !context->clear_on_transition) {
+		obs_frontend_add_event_callback(ds_frontend_event, data);
+		context->clear_on_transition = clear_on_transition;
+	} else if (!clear_on_transition && context->clear_on_transition){
+		obs_frontend_remove_event_callback(ds_frontend_event, data);
+		context->clear_on_transition = clear_on_transition;
+	}
 	context->max_undo = (uint32_t)obs_data_get_int(settings, "max_undo");
 	context->size.x = (float)obs_data_get_int(settings, "width");
 	context->size.y = (float)obs_data_get_int(settings, "height");
 	context->tool = (uint32_t)obs_data_get_int(settings, "tool");
 	context->show_mouse = obs_data_get_bool(settings, "show_cursor");
-	context->cursor_size = (float)obs_data_get_double(settings, "cursor_size");
+	context->cursor_size = obs_data_get_bool(settings, "cursor_custom_size") ? (float)obs_data_get_double(settings, "cursor_size") : 0.0f;
 	vec4_from_rgba(&context->cursor_color, (uint32_t)obs_data_get_int(settings, "cursor_color"));
 	context->cursor_color.w = 1.0f;
 	vec4_from_rgba(&context->tool_color, (uint32_t)obs_data_get_int(settings, "tool_color"));
@@ -662,7 +666,7 @@ static bool clear_property_button(obs_properties_t *props, obs_property_t *prope
 	UNUSED_PARAMETER(props);
 	UNUSED_PARAMETER(property);
 	struct draw_source *ds = data;
-	clear(ds);
+	draw_clear(ds);
 	return false;
 }
 
@@ -714,12 +718,20 @@ static obs_properties_t *ds_get_properties(void *data)
 	obs_property_float_set_suffix(p, "px");
 
 	obs_properties_add_color(props, "cursor_color", obs_module_text("CursorColor"));
-	p = obs_properties_add_float_slider(props, "cursor_size", obs_module_text("CursorSize"), 0.0, 100.0, 0.1);
+
+	obs_properties_t *cursor = obs_properties_create();
+
+	p = obs_properties_add_float_slider(cursor, "cursor_size", obs_module_text("CursorSize"), 0.0, 100.0, 0.1);
 	obs_property_float_set_suffix(p, "px");
+
+	obs_properties_add_group(props, "cursor_custom_size", obs_module_text("CursorCustomSize"), OBS_GROUP_CHECKABLE, cursor);
+
 	obs_properties_add_path(props, "cursor_file", obs_module_text("CursorFile"), OBS_PATH_FILE, image_filter, NULL);
 
 	obs_properties_add_int(props, "max_undo", obs_module_text("UndoMax"), 1, 10000, 1);
 
+	obs_properties_add_bool(props, "clear_on_scene_transition", obs_module_text("ClearOnSceneTransition"));
+
 	obs_properties_add_button2(props, "clear", obs_module_text("Clear"), clear_property_button, data);
 
 	obs_properties_add_text(props, "plugin_info",
@@ -739,6 +751,7 @@ static void ds_get_defaults(obs_data_t *settings)
 	obs_data_set_default_int(settings, "tool_color", 0xFF0000FF);
 	obs_data_set_default_double(settings, "tool_alpha", 100.0);
 	obs_data_set_default_bool(settings, "show_cursor", true);
+	obs_data_set_default_bool(settings, "custom_cursor_size", true);
 	obs_data_set_default_double(settings, "cursor_size", 10);
 	obs_data_set_default_int(settings, "max_undo", 10);
 }
